% plot runtime

clear;
clc;
close all;

%% load data

mSwpOrd = dlmread('runtime_swp_ord.txt',' ');
mSwpMem = dlmread('runtime_swp_mem.txt',' ');

%% separate columns

vOrd = mSwpOrd(:,1); 
vMem = mSwpMem(:,2);

nOrd = mSwpMem(1,1);
nMem = mSwpOrd(1,2);

mSwpOrdComb  = mSwpOrd(:,4:7); 
mSwpOrdTable = mSwpOrd(:,8:11); 
mSwpOrdLoop  = mSwpOrd(:,12:15);
mSwpOrdHard  = mSwpOrd(:,16:19);

mSwpMemComb  = mSwpMem(:,4:7); 
mSwpMemTable = mSwpMem(:,8:11); 
mSwpMemLoop  = mSwpMem(:,12:15);
mSwpMemHard  = mSwpMem(:,16:19);

%% create plot

figure;
set(gcf,'Position',[500,500,1200,300]);

mColor = [ ...
  0.000,  0.447,  0.741
  0.850,  0.325,  0.098
  0.929,  0.694,  0.125
  0.494,  0.184,  0.556
  0.466,  0.674,  0.188
  0.301,  0.745,  0.933
  0.635,  0.078,  0.184
  0,0,0
];

vLimOrd = [vOrd(1),vOrd(end)]; 
vLimMem = [vMem(1),vMem(end)];

vTickOrd = (vOrd(1):vOrd(end)).';
vTickMem = (vMem(1):vMem(end)).';

mAxes  = zeros(2,4);
mLines = zeros(2,4,4);

mAxes(1,1)    = subplot(2,4,1);
mLines(1,1,:) = plot(vOrd,mSwpOrdComb);
grid on;
title('Combinatoric','Interpreter','latex');
xlabel(['Order $\vert$ Memory = ',num2str(nMem)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimOrd);
set(gca,'XTick',vTickOrd);

mAxes(1,2)    = subplot(2,4,2);
mLines(1,2,:) = plot(vOrd,mSwpOrdTable);
grid on;
title('Lookup Table','Interpreter','latex');
xlabel(['Order $\vert$ Memory = ',num2str(nMem)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimOrd);
set(gca,'XTick',vTickOrd);

mAxes(1,3)    = subplot(2,4,3);
mLines(1,3,:) = plot(vOrd,mSwpOrdLoop);
grid on;
title('Nested Loop','Interpreter','latex');
xlabel(['Order $\vert$ Memory = ',num2str(nMem)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimOrd);
set(gca,'XTick',vTickOrd);

mAxes(1,4)    = subplot(2,4,4);
mLines(1,4,:) = plot(vOrd,mSwpOrdHard);
grid on;
title('Hard Coded','Interpreter','latex');
xlabel(['Order $\vert$ Memory = ',num2str(nMem)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimOrd);
set(gca,'XTick',vTickOrd);

mAxes(2,1)    = subplot(2,4,5);
mLines(2,1,:) = plot(vMem,mSwpMemComb);
grid on;
xlabel(['Memory $\vert$ Order = ',num2str(nOrd)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimMem);
set(gca,'XTick',vTickMem);

mAxes(2,2)    = subplot(2,4,6);
mLines(2,2,:) = plot(vMem,mSwpMemTable);
xlabel(['Memory $\vert$ Order = ',num2str(nOrd)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimMem);
grid on;
set(gca,'XTick',vTickMem);

mAxes(2,3)    = subplot(2,4,7);
mLines(2,3,:) = plot(vMem,mSwpMemLoop);
grid on;
xlabel(['Memory $\vert$ Order = ',num2str(nOrd)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimMem);
set(gca,'XTick',vTickMem);

mAxes(2,4)    = subplot(2,4,8);
mLines(2,4,:) = plot(vMem,mSwpMemHard);
grid on;
xlabel(['Memory $\vert$ Order = ',num2str(nOrd)],'Interpreter','latex');
ylabel('$\mu$s / sample','Interpreter','latex');
xlim(vLimMem);
set(gca,'XTick',vTickMem);

set(mLines, 'LineWidth', 2);
set(mLines, 'MarkerSize', 8);
for iRow = 1 : 2  
  for iCol = 1 : 4    
    set(mLines(iRow,iCol,1), 'Color', mColor(1,:), 'Marker', 'x');    
    set(mLines(iRow,iCol,2), 'Color', mColor(5,:), 'Marker', 'x');
    set(mLines(iRow,iCol,3), 'Color', mColor(1,:), 'Marker', 'o');
    set(mLines(iRow,iCol,4), 'Color', mColor(5,:), 'Marker', 'o');
  end
end

hLeg  = legend('Direct 1','Reuse','Direct 2','Horner','Location','NorthWest');
set(hLeg,'Position',[0.875 0.18 0.09 0.76],'Interpreter','latex','FontSize',11,'Color','none','Box','off');

hLeg0 = annotation('rectangle',[0.8731 0.1567 0.103 0.7689],'Color','k');
hLeg1 = annotation('textbox',[0.9125,0.706,0.1,0.1],'String','$y[n]$ $\Phi_i[n]$','Interpreter','latex','EdgeColor','none','FontSize',11);
hLeg2 = annotation('textbox',[0.9125,0.529,0.1,0.1],'String','$y[n]$ $\Phi_i[n]$','Interpreter','latex','EdgeColor','none','FontSize',11);
hLeg3 = annotation('textbox',[0.9125,0.349,0.1,0.1],'String','$y[n]$ only','Interpreter','latex','EdgeColor','none','FontSize',11);
hLeg4 = annotation('textbox',[0.9125,0.169,0.1,0.1],'String','$y[n]$ only','Interpreter','latex','EdgeColor','none','FontSize',11);

pause(0.5);

for iRow = 1 : 2  
  for iCol = 1 : 4
    hAxis = mAxes(iRow,iCol);
    set(hAxis,'Position',get(hAxis,'Position')+[-0.055,0,0,0]);
  end
end
