#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "volterra.h"

// number of evaluation runs implemented in volterraEvalRun()
#define NR 16

int *C, *m, *tab;
smp *x, *y, **yi, *h, *xx, *g;
double *t;

void generateRandomSamples(smp *x, int N) {
  int n;
  for (n=0; n<N; n++) {
    x[n] = (smp) rand() / RAND_MAX - 0.5;
  }
}

void volterraEvalInit(int P, int M, int N) {
  int i;
  C   = calloc(P+1,sizeof(int));
  m   = calloc(P+1,sizeof(int));  
  volterraSize(P,M,C);
  tab = calloc((P+1)*C[P],sizeof(int));
  x   = calloc(N+M,sizeof(smp));  
  y   = calloc(NR*N,sizeof(smp));
  yi  = calloc(NR,sizeof(smp*));
  h   = calloc(C[P],sizeof(smp));
  xx  = calloc(C[P],sizeof(smp));
  g   = calloc(C[P-1],sizeof(smp));
  t   = calloc(NR,sizeof(double));  
  x += M;
  for (i=0; i<NR; i++) {
    yi[i] = y + i*N;
  }
  generateRandomSamples(x, N);
  generateRandomSamples(h, C[P]);
}

void volterraEvalRun(int P, int M, int N, int R) {  
  double tic, toc;
  int r, n, i = 0;
  smp *px, *py, *px0 = x, *py0;  
  void (*volterraFilterDirect1Hard)(smp*, smp*, smp*, int, smp*) = NULL;
  void (*volterraFilterDirect2Hard)(smp*, smp*, smp*, int) = NULL;  
  void (*volterraFilterReuseHard)(smp*, smp*, smp*, int, smp*) = NULL;
  void (*volterraFilterHornerHard)(smp*, smp*, smp*, int, smp*) = NULL;    
  volterraSize(P,M,C);
  volterraTable(tab,P,M,m);
  if (R < 0) {
    R = -R / (N * C[P]);  
    if (R == 0) R = 1;
  }
  // Comb - Direct1
  py0 = yi[i];
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterDirect1Comb(px++, py++, h, P, M, xx, m);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);
  // Comb - Reuse
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterReuseComb(px++, py++, h, P, M, xx, m);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);   
  // Comb - Direct2
  py0 = yi[i];
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterDirect2Comb(px++, py++, h, P, M, m);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);   
  // Comb - Horner
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterHornerComb(px++, py++, h, P, M, C, g, m);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);  
  // Table - Direct1 
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterDirect1Table(px++, py++, h, P, C, xx, tab);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);  
  // Table - Reuse
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterReuseTable(px++, py++, h, P, M, C, xx, tab);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);    
  // Table - Direct2
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterDirect2Table(px++, py++, h, P, C, tab);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);    
  // Table - Horner
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterHornerTable(px++, py++, h, P, M, C, g, tab);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);  
  // Loop - Direct1
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterDirect1Loop(px++, py++, h, P, M, xx);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);  
  // Loop - Reuse
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterReuseLoop(px++, py++, h, P, M, xx);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);    
  // Loop - Direct2
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterDirect2Loop(px++, py++, h, P, M);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);
  // Loop - Horner
  py0 = yi[i];  
  tic = clock();
  for (r=1; r<=R; r++) {
    px = px0;
    py = py0;  
    for (n=0; n<N; n++) {
      volterraFilterHornerLoop(px++, py++, h, P, M, C, g);
    }
  }
  toc = clock();
  t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);  
  // Hard - Select Implementation
  switch (M) {
    case 0:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem0;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem0;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem0;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem0;
      break;
    case 1:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem1;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem1;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem1;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem1;      
      break;
    case 2:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem2;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem2;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem2;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem2;
      break;
    case 3:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem3;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem3;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem3;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem3;      
      break;
    case 4:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem4;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem4;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem4;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem4;            
      break;
    case 5:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem5;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem5;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem5;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem5;
      break;
    case 6:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem6;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem6;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem6;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem6;
      break;
    case 7:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem7;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem7;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem7;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem7;
      break;
    case 8:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem8;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem8;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem8;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem8;
      break;
    case 9:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem9;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem9;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem9;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem9;
      break;
    case 10:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem10;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem10;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem10;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem10;
      break;
    case 11:
      volterraFilterDirect1Hard = &volterraFilterDirect1HardMem11;
      volterraFilterDirect2Hard = &volterraFilterDirect2HardMem11;      
      volterraFilterReuseHard   = &volterraFilterReuseHardMem11;
      volterraFilterHornerHard  = &volterraFilterHornerHardMem11;
      break;            
    default:
      fprintf(stderr, "There is no hard implementation for memory depth %d.\n", M);
  }  
  // Hard - Direct1
  if (volterraFilterDirect1Hard != NULL) {
    py0 = yi[i];    
    tic = clock();
    for (r=1; r<=R; r++) {
      px = px0;
      py = py0;  
      for (n=0; n<N; n++) {
        (*volterraFilterDirect1Hard)(px++, py++, h, P, xx);
      }
    }
    toc = clock();
    t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);
  } else {
    t[i++] = 0;
  }
  // Hard - Reuse
  if (volterraFilterReuseHard != NULL) {
    py0 = yi[i];        
    tic = clock();
    for (r=1; r<=R; r++) {
      px = px0;
      py = py0;  
      for (n=0; n<N; n++) {
        (*volterraFilterReuseHard)(px++, py++, h, P, xx);
      }
    }
    toc = clock();
    t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);   
  } else {
    t[i++] = 0;
  }  
  // Hard - Direct2
  if (volterraFilterDirect2Hard != NULL) {
    py0 = yi[i];    
    tic = clock();
    for (r=1; r<=R; r++) {
      px = px0;
      py = py0;  
      for (n=0; n<N; n++) {
        (*volterraFilterDirect2Hard)(px++, py++, h, P);
      }
    }
    toc = clock();
    t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);
  } else {
    t[i++] = 0;
  }
  // Hard - Horner
  if (volterraFilterHornerHard != NULL) {
    py0 = yi[i];        
    tic = clock();
    for (r=1; r<=R; r++) {
      px = px0;
      py = py0;  
      for (n=0; n<N; n++) {
        (*volterraFilterHornerHard)(px++, py++, h, P, g);
      }
    }
    toc = clock();
    t[i++] = (double)(toc-tic)/(CLOCKS_PER_SEC*1e-6*R*N);  
  } else {
    t[i++] = 0;
  }
}

void volterraEvalPrintSignals(int N) {
  int i, n;
  printf("Evaluation of correctness by comparison of output signals y[n]\n\n");
  printf("    ");
  printf("Comb  ------------------------- ");
  printf("Table ------------------------- ");
  printf("Loop  ------------------------- ");  
  printf("Hard  ------------------------- ");
  printf("\n");
  printf("  n ");  
  printf("Direct1   Reuse Direct2  Horner ");
  printf("Direct1   Reuse Direct2  Horner ");
  printf("Direct1   Reuse Direct2  Horner ");
  printf("Direct1   Reuse Direct2  Horner ");    
  printf("\n");  
  for (n=0; n<N; n++) {
    printf("%3d ", n);
    for (i=0; i<NR; i++) {
      printf("%+7.3f ", yi[i][n]);
    }
    printf("\n");
  }
}

void volterraEvalPrintRuntimeHeader(int Pmin, int Pmax, int Mmin, int Mmax) {
  printf("Runtime in us/sample");
  printf(", P = maximum order");  
  printf(", M = memory depth");  
  printf(", C = number of parameters");      
  printf(", Sweep: Pmin=%d Pmax=%d Mmin=%d Mmax=%d \n\n", Pmin, Pmax, Mmin, Mmax);  
  printf("Setup ----- ");
  printf("Comb  ------------------------- ");
  printf("Table ------------------------- ");
  printf("Loop  ------------------------- ");  
  printf("Hard  ------------------------- ");
  printf("\n");
  printf("  P   M   C ");
  printf("Direct1   Reuse Direct2  Horner ");
  printf("Direct1   Reuse Direct2  Horner ");
  printf("Direct1   Reuse Direct2  Horner ");
  printf("Direct1   Reuse Direct2  Horner ");
  printf("\n");
}

void volterraEvalPrintRuntime(int P, int M) {
  int i;
  printf("%3d ", P);
  printf("%3d ", M);
  printf("%3d ", C[P]);    
  for (i=0; i<NR-1; i++) {
    printf("%7.3f ", t[i]);
  }
  printf("%7.3f\n", t[i]);
}

void volterraEvalStoreRuntime(int P, int M, FILE *fp) {
  int i;
  fprintf(fp, "%03d ", P);
  fprintf(fp, "%03d ", M);
  fprintf(fp, "%03d ", C[P]);
  for (i=0; i<NR-1; i++) {
    fprintf(fp, "%07.3f ", t[i]);
  }
  fprintf(fp, "%07.3f\n", t[i]);
}
